from collections import defaultdict
from itertools import chain


def flatten(iterable):
    return list(chain.from_iterable(iterable))


def leaders(iterable, top=10):
    counters = defaultdict(int)
    for item in iterable:
        counters[item] += 1

    return sorted(list(counters.items()), reverse=True, key=lambda tup: tup[1])[:top]



# Python imports
import bz2
import gzip
import logging
import os
import posix
import stat
import tarfile
import time
import zipfile
import warnings

# Third party imports
try:
    import lzma
except ImportError:
    try:
        import backports.lzma as lzma
    except ImportError:
        lzma = None
        warnings.warn('Could not import optional lzma module')
try:
    import rarfile
except ImportError:
    rarfile = None
    warnings.warn('Could not import optional rarfile module')

# Project imports
try:
    from classified import magic
except ImportError:
    magic = None  # pyflakes.ignore
    warnings.warn('Could not import required python-magic module')
    raise
# Project imports (platform dependant)
from classified.platform import get_filesystem, get_filesystems


class CorruptionError(ValueError):
    pass


class Path(object):
    def __init__(self, path, parent=None):
        self.path = os.path.abspath(path)
        self.parent = parent

        # Flags used by recursor
        self.walkable = True
        self.readable = False

        # Normalise path
        if os.path.islink(self.path):
            result = os.readlink(self.path)
            if result.startswith(os.sep):
                self.path = os.path.abspath(result)
            else:
                self.path = os.path.abspath(os.path.join(
                    os.path.dirname(self.path), result
                ))

        # Proxy attributes
        for attr in dir(self.path):
            if attr.startswith('_'):
                continue

            elif hasattr(self, attr):
                continue

            else:
                setattr(self, attr, getattr(self.path, attr))

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.path)

    def __str__(self):
        return self.path

    def probe(cls, path):
        try:
            fileinfo = os.stat(path)
            filemode = fileinfo[stat.ST_MODE]
        except (IOError, OSError):
            return False
        else:
            return stat.S_ISDIR(filemode)

    probe = classmethod(probe)

    def walk(self, recurse=True, depth=0, max_depth=10,
             deflate=True, deflate_limit=0):

        logging.debug('%s depth %d/%d' % (self.path, depth, max_depth))

        for item in self.walk_tree(deflate, deflate_limit):
            if item is None:
                continue

            else:
                yield item

                if recurse and (max_depth and depth >= max_depth):
                    logging.warning('%s max recursion depth' % (self.path,))
                    continue

                if recurse and item.walkable:
                    try:
                        for sub in item.walk(depth=depth + 1, max_depth=max_depth):
                            yield sub
                    except (IOError, OSError) as error:
                        logging.error('%s error %s' % (self.path, str(error)))
                    except CorruptionError as error:
                        logging.error('%s is corrupt' % (self.path,))

    def walk_tree(self, deflate, deflate_limit):
        if os.access(self.path, os.R_OK):
            for item in os.listdir(self.path):
                full = os.path.join(self.path, item)
                if Path.probe(full):
                    yield Path(full, parent=self)
                else:
                    mount_hint = getattr(self, '_mount', None)
                    yield File.maybe(
                        full,
                        deflate_if_archive=deflate,
                        deflate_limit=deflate_limit,
                        mount_hint=mount_hint,
                        parent=self,
                    )
        else:
            logging.error('%s not accessible, skipping' % (self.path,))


class Repository(Path):
    supported_types = {
        # GNU Arch
        'arch': (
            ('file', os.path.join('{arch}', '.arch-project-tree')),
        ),
        # Bazaar
        'bzr': (
            ('path', os.path.join('.bzr', 'repository')),
        ),
        # Concurrent Versions System
        'cvs': (
            ('path', 'CVS'),
            ('path', 'CVSROOT'),
        ),
        # DARCS
        'darcs': (
            ('path', os.path.join('_darcs', 'pristine.hashed')),
        ),
        # Git SCM
        'git': (
            ('path', os.path.join('.git', 'objects')),      # repository
            ('path', os.path.join('refs', 'heads')),        # bare repository
        ),
        # Mercurial
        'hg': (
            ('path', os.path.join('.hg', 'store')),
        ),
        # Monotone
        'monotone': (
            ('file', os.path.join('_MTN', 'format')),
        ),
        # Revision Control System
        'rcs': (
            ('path', 'RCS'),
        ),
        # Subversion (SVN)
        'svn': (
            ('file', os.path.join('.svn', 'format')),   # repository (checkout)
            ('path', os.path.join('db', 'revs')),       # repository
        ),
    }
    type_cache = {}

    def __init__(self, path):
        super(Repository, self).__init__(path)
        self.type = self._detect_type()

    def _detect_type(self):
        node = self
        while node is not None:
            # This only works for directories
            if not isinstance(node, Path):
                node = node.parent or Path(os.path.dirname(node.path))
            try:
                return Repository.type_cache[node.path]
            except KeyError:
                repository_type = self._detect_type_path(node.path)

            if repository_type is None:
                if node.parent is None:
                    path = os.path.dirname(node.path)
                    # Stop iterating if we hit the file system root
                    if path != node.path and os.path.isdir(path):
                        node = Path(path)
                    else:
                        node = None
                else:
                    node = node.parent
            else:
                Repository.type_cache[node.path] = repository_type
                return repository_type

    def _detect_type_path(self, path):
        for vendor, probes in self.supported_types.items():
            for filetype, filename in probes:
                filepath = os.path.join(path, filename)
                try:
                    fileinfo = os.stat(filepath)
                    filemode = fileinfo[stat.ST_MODE]
                except (IOError, OSError):
                    continue

                # Common types
                if filetype == 'path' and stat.S_ISDIR(filemode):
                    return vendor
                if filetype == 'file' and stat.S_ISREG(filemode):
                    return vendor
                # Less common types
                if filetype == 'char' and stat.S_ISCHR(filemode):
                    return vendor
                if filetype == 'fifo' and stat.S_ISFIFO(filemode):
                    return vendor
                if filetype == 'link' and stat.S_ISLNK(filemode):
                    return vendor
                if filetype == 'sock' and stat.S_ISSOCK(filemode):
                    return vendor

        # It's not a repository (that we know of)
        return None


class File(Path):
    Corrupt = CorruptionError

    def __init__(self, path, parent):
        super(File, self).__init__(path, parent=parent)

        self.handle = None

        # Clone?
        if isinstance(path, File):
            self.walkable = path.walkable
            self.readable = path.readable
            self.mimetype = path.mimetype

        else:
            # Flags used by recursor
            self.walkable = False
            self.readable = True

    def __repr__(self):
        return '<%s %s mime-type=%s fs-type=%s repository-type=%s>' % (
            self.__class__.__name__,
            self.path,
            self.mimetype or 'unknown',
            self.mount.fs['type'],
            self.repository.type
        )

    # Method that does archive detection
    def maybe(path, deflate_if_archive=True, deflate_limit=0, mount_hint=None, parent=None):
        instance = File(path, parent=parent)
        if deflate_if_archive and instance.mimetype in Archive.supported_mimetypes:
            if deflate_limit > 0 and instance.size > deflate_limit:
                logging.warning('skipped archive %s: too big (%s > %s)' % \
                    (instance, instance.size, deflate_limit))
                return instance

            try:
                instance = Archive(path, mount_hint, parent=parent)
                logging.debug('opened archive %s: %s' % (instance,
                    instance.mimetype))
            except CorruptionError as e:
                logging.warn('failed to inspect archive %s: %s' % (instance,
                    e))

        return instance

    maybe = staticmethod(maybe)

    # Low-level file like methods

    def __iter__(self):
        return iter(self.handle)

    def close(self):
        self.handle.close()
        return self

    def read(self, size=None):
        try:
            return self.handle.read(size)
        except (EOFError, IOError):
            raise self.__class__.Corrupt(self.path)

    def open(self, mode='r'):
        self.handle = open(self.path, mode)
        return self

    def seek(self, offset, whence=0):
        self.handle.seek(offset, whence)
        return self

    def stat(self):
        return os.stat(self.path)

    def tell(self):
        return self.handle.tell()

    def readline(self):
        return self.handle.readline()

    def readlines(self):
        return self.handle.readlines()

    # Dynamic properties

    def mimetype_get(self):
        if not hasattr(self, '_mimetype'):
            try:
                self._mimetype = magic.from_file(self.path, mime=True)
                self._mimetype = self._mimetype.decode('ascii')
            except NameError:
                self._mimetype = None
        return self._mimetype

    def mimetype_set(self, mimetype):
        self._mimetype = mimetype

    mimetype = property(mimetype_get, mimetype_set)

    def mount_get(self):
        if not hasattr(self, '_mount'):
            self._mount = Mount(self.path)
        return self._mount

    def mount_set(self, mount):
        self._mount = mount

    mount = property(mount_get, mount_set)

    def mtime_get(self):
        return self.stat()[stat.ST_MTIME]

    mtime = property(mtime_get)

    def repository_get(self):
        if not hasattr(self, '_repository'):
            self._repository = Repository(self.path)
        return self._repository

    repository = property(repository_get)

    def size_get(self):
        return self.stat()[stat.ST_SIZE]

    size = property(size_get)


class Mount(Path):
    def __init__(self, path):
        super(Mount, self).__init__(path)
        self.fs = self._detect_fs()

    def _detect_fs(self):
        if not hasattr(Mount, '_fs_cache') or \
                Mount._fs_cache_timeout < time.time():
            setattr(Mount, '_fs_cache', list(get_filesystems()))
            setattr(Mount, '_fs_cache_timeout', time.time() + 60)

        return get_filesystem(str(self), Mount._fs_cache)


class Archive(File):
    supported_mimetypes = [
        'application/x-bzip2',
        'application/x-gzip',
        'application/x-tar',
        'application/zip',
    ]

    def __init__(self, path, mount_hint=None, parent=None):
        super(Archive, self).__init__(path, parent=parent)

        # Flags used by recursor
        self.walkable = True
        self.readable = False

        if not mount_hint is None:
            self._mount = mount_hint

        base, mimetype = self.mimetype.split('/', 1)
        if mimetype in ['x-bzip2', 'x-gzip', 'x-xz']:
            # Try to see if this is a compressed file, or a compressed archive
            self.bundle = tarfile.is_tarfile(self.path)
            if self.bundle:
                self.recursor = self._recursor_tar
                try:
                    self.handle = tarfile.open(self.path)
                except tarfile.ReadError:
                    raise Archive.Corrupt(self.path)
            else:
                self.recursor = self._recursor_compressed

                if mimetype == 'x-bzip2':
                    self.handle = bz2.BZ2File(self.path)
                elif mimetype == 'x-gzip':
                    self.handle = gzip.open(self.path)
                elif mimetype == 'x-xz':
                    # Both lzma, backports.lzma and pyliblzma provide the lzma
                    # module for importing, yet they have a different API
                    if hasattr(lzma, 'open'):
                        self.handle = lzma.open(self.path)
                    elif hasattr(lzma, 'LZMAFile'):  # pyliblzma
                        self.handle = lzma.LZMAFile(self.path, mode='r')

                # Override mimetype by the mimetype of the compressed file
                self.mimetype = magic.from_buffer(self.read(1024), mime=True)
                self.mimetype = self.mimetype.decode('ascii')

        elif mimetype == 'x-rar' and rarfile is not None:
            try:
                self.handle = rarfile.RarFile(self.path)
            except OSError:
                warnings.userwarning('Failed to open rar archive, have you '
                                     'installed the unrar binary?')
            self.recursor = self._recursor_rar

        elif mimetype == 'x-tar':
            try:
                self.handle = tarfile.open(self.path)
            except tarfile.ReadError:
                raise Archive.Corrupt(self.path)
            self.recursor = self._recursor_tar

        elif mimetype == 'zip':
            self.recursor = self._recursor_zip
            self.handle = zipfile.ZipFile(self.path)

    def _recursor_compressed(self, depth, max_depth):
        yield ArchiveFile(self.path, self)

    def _recursor_rar(self, depth, max_depth):
        for item in self.handle.infolist():
            size = len(item.filename.split(os.sep))
            if (depth + size - 1) >= max_depth:
                continue
            full = os.path.join(self.path, item.filename)
            try:
                yield ArchiveFile(full, self)
            except KeyError:  # File not in archive
                pass

    def _recursor_tar(self, depth, max_depth):
        try:
            for item in self.handle.getmembers():
                size = len(item.name.split(os.sep))
                if (depth + size - 1) >= max_depth:
                    continue
                full = os.path.join(self.path, item.name)
                if item.type == tarfile.REGTYPE:
                    try:
                        yield ArchiveFile(full, self)
                    except KeyError:  # File not in archive
                        pass
        except tarfile.ReadError:
            raise Archive.Corrupt(self.path)

    def _recursor_zip(self, depth, max_depth):
        for item in self.handle.infolist():
            size = len(item.filename.split(os.sep))
            if (depth + size - 1) >= max_depth:
                continue
            full = os.path.join(self.path, item.filename)
            try:
                yield ArchiveFile(full, self)
            except KeyError:  # File not in archive
                pass

    def walk(self, depth=0, max_depth=10):
        if self.walkable and (not max_depth or depth < max_depth):
            for item in self.recursor(depth=depth+1, max_depth=max_depth):
                yield item


class ArchiveFile(File):
    def __init__(self, path, archive):
        super(ArchiveFile, self).__init__(path, parent=archive)

        # Archive that contains this file
        self.archive = archive

        # Name of this file in the archive
        self.filename = self.path.replace(self.archive.path, '')
        self.filename = self.filename.lstrip(os.sep)

        self.walkable = False

        if rarfile and isinstance(self.archive.handle, rarfile.RarFile):
            self.member = self.archive.handle.getinfo(self.filename)
            self.readable = True

        elif isinstance(self.archive.handle, tarfile.TarFile):
            self.member = self.archive.handle.getmember(self.filename)
            self.readable = self.member.isreg()

        elif isinstance(self.archive.handle, zipfile.ZipFile):
            self.member = self.archive.handle.getinfo(self.filename)
            self.readable = True

    def open(self, mode='r'):
        if 'w' in mode or 'a' in mode:
            raise IOError('Archives are read-only')
        if not self.readable:
            raise IOError('Archive member not readable')

        if isinstance(self.archive.handle, bz2.BZ2File):
            try:
                self.handle = bz2.BZ2File(self.archive.path, mode=mode)
                return self.handle
            except IOError as e:
                raise Archive.Corrupt(self.path)

        elif isinstance(self.archive.handle, gzip.GzipFile):
            try:
                self.handle = gzip.GzipFile(self.archive.path, mode=mode)
                return self.handle
            except IOError as e:
                raise Archive.Corrupt(self.path)

        elif lzma and isinstance(self.archive.handle, lzma.LZMAFile):
            try:
                self.handle = lzma.LZMAFile(self.archive.path, mode=mode)
                return self.handle
            except lzma.LZMAError as e:
                raise Archive.Corrupt(self.path)

        elif rarfile and isinstance(self.archive.handle, rarfile.RarFile):
            mode = mode.replace('b', '')  # not supported in rar files
            try:
                self.handle = self.archive.handle.open(self.filename, mode)
                return self.handle
            except rarfile.Error as e:
                logging.warning(str(e))
                raise Archive.Corrupt(self.path)
            except OSError:
                logging.error('failed to open rar archive, did you install '
                              'the unrar binary?')
                raise Archive.Corrupt(self.path)
            except TypeError as e:
                logging.error('failed to open rar archive, file corrupt? %s'
                              % str(e))
                raise Archive.Corrupt(self.path)

        elif isinstance(self.archive.handle, tarfile.TarFile):
            try:
                self.handle = tarfile.TarFile.fileobject(self.archive.handle,
                    self.member)
                return self.handle
            except tarfile.TarError as e:
                logging.info(str(e))
                raise Archive.Corrupt(self.path)

        elif isinstance(self.archive.handle, zipfile.ZipFile):
            try:
                mode = mode.replace('b', '')  # not supported in zip files
                self.handle = self.archive.handle.open(self.filename, mode)
                return self.handle
            except zipfile.BadZipfile as e:
                logging.warning(str(e))
                raise Archive.Corrupt(self.path)
            except RuntimeError as e:
                error = str(e)
                if 'password required' in error:
                    logging.info(error)
                else:
                    logging.critical(error)
                raise Archive.Corrupt(self.path)

        else:
            raise NotImplementedError('Archive format not supported')

    def stat(self):
        '''
        Provides an interface like ``os.stat()``
        '''
        if isinstance(self.archive.handle, bz2.BZ2File):
            return self.archive.stat()

        elif isinstance(self.archive.handle, gzip.GzipFile):
            return self.archive.stat()

        elif lzma and isinstance(self.archive.handle, lzma.LZMAFile):
            return self.archive.stat()

        elif isinstance(self.archive.handle, tarfile.TarFile):
            return posix.stat_result((
                self.member.mode,
                -1,
                -1,
                -1,
                self.member.uid,
                self.member.gid,
                self.member.size,
                self.member.mtime,
                self.member.mtime,
                self.member.mtime,
            ))

        elif rarfile and isinstance(self.archive.handle, rarfile.RarFile):
            info = self.archive.stat()
            return posix.stat_result((
                info[stat.ST_MODE],
                -1,
                -1,
                -1,
                info[stat.ST_UID],
                info[stat.ST_GID],
                self.member.file_size,
                time.mktime(self.member.date_time + (0, 0, 0)),
                time.mktime(self.member.date_time + (0, 0, 0)),
                time.mktime(self.member.date_time + (0, 0, 0)),
            ))

        elif isinstance(self.archive.handle, zipfile.ZipFile):
            info = self.archive.stat()
            return posix.stat_result((
                info[stat.ST_MODE],
                -1,
                -1,
                -1,
                info[stat.ST_UID],
                info[stat.ST_GID],
                self.member.file_size,
                time.mktime(self.member.date_time + (0, 0, 0)),
                time.mktime(self.member.date_time + (0, 0, 0)),
                time.mktime(self.member.date_time + (0, 0, 0)),
            ))

    def mimetype_get(self):
        if not hasattr(self, '_mimetype'):
            try:
                self.open('rb')
            except CorruptionError:
                self._mimetype = None
            else:
                self._mimetype = magic.from_buffer(self.read(1024), mime=True)
                self._mimetype = self._mimetype.decode('ascii')
        return self._mimetype

    mimetype = property(mimetype_get)


# Support for these mime types depend on the availability of third party
# modules (or Python 3.x for some)
if lzma is not None:
    Archive.supported_mimetypes.append('application/x-xz')
if rarfile is not None:
    Archive.supported_mimetypes.append('application/x-rar')


if __name__ == '__main__':
    path = File(os.path.expanduser('~'))
    print(repr(path))



# Python imports
import os
import re


# Project imports
from classified.probe.base import Probe


class Password(Probe):
    default_pattern = r'\bpass(?:|wd|word)\b[ \s\t=:]+(?P<password>.*)'
    target = ('text/*',)
    format = '{filename_relative}[{line}]: {type} {text_masked}'

    def __init__(self, config, *args, **kwargs):
        super(Password, self).__init__(config, *args, **kwargs)

        try:
            pattern = self.config.get('probe:password', 'pattern')
        except self.config.NoOptionError:
            pattern = self.default_pattern
        if isinstance(pattern, bytes):
            pattern = pattern.decode('ascii')
        self.re_password = re.compile(pattern)

    def probe(self, item):
        filename = str(item)
        basename = os.path.basename(filename)

        item.open()
        try:
            if 'pgpass' in basename:
                self.probe_pgpass(item)

            # Scan for passwords heuristically in any case            
            line = 0
            for text in item:
                if isinstance(text, bytes):
                    text = text.decode('utf-8')
                line += 1
                for password in self.re_password.findall(text):
                    if password == '':
                        continue

                    self.record(item,
                        raw=text,
                        type='password',
                        line=line,
                        text=text.rstrip(),
                        text_masked=text.rstrip().replace(password, '********'),
                        password=password,
                        password_masked='********',
                    )
        finally:
            item.close()

    def probe_pgpass(self, item):
        line = 0
        for text in item:
            line += 1
            part = text.split(':')
            if len(part) == 5 and part[4]:
                self.record(item,
                    raw=text,
                    type='pgpass',
                    line=line,
                    text=text.rstrip(),
                    text_masked=':'.join(part[:4] + ['********']),
                    password=part[4],
                    password_masked='********',
                )

        item.seek(0)  # reset file handle



"""
magic is a wrapper around the libmagic file identification library.

See README for more information.

Usage:

>>> import magic
>>> magic.from_file("testdata/test.pdf")
'PDF document, version 1.2'
>>> magic.from_file("testdata/test.pdf", mime=True)
'application/pdf'
>>> magic.from_buffer(open("testdata/test.pdf").read(1024))
'PDF document, version 1.2'
>>>


"""

import sys
import glob
import os.path
import ctypes
import ctypes.util
import threading

from ctypes import c_char_p, c_int, c_size_t, c_void_p

class MagicException(Exception): pass

class Magic:
    """
    Magic is a wrapper around the libmagic C library.

    """

    def __init__(self, mime=False, magic_file=None, mime_encoding=False,
                 keep_going=False):
        """
        Create a new libmagic wrapper.

        mime - if True, mimetypes are returned instead of textual descriptions
        mime_encoding - if True, codec is returned
        magic_file - use a mime database other than the system default
        keep_going - don't stop at the first match, keep going
        """
        self.flags = MAGIC_NONE
        if mime:
            self.flags |= MAGIC_MIME
        elif mime_encoding:
            self.flags |= MAGIC_MIME_ENCODING
        if keep_going:
            self.flags |= MAGIC_CONTINUE

        self.cookie = magic_open(self.flags)

        magic_load(self.cookie, magic_file)

        self.thread = threading.currentThread()

    def from_buffer(self, buf):
        """
        Identify the contents of `buf`
        """
        self._thread_check()
        try:
            return magic_buffer(self.cookie, buf)
        except MagicException as e:
            return self._handle509Bug(e)

    def from_file(self, filename):
        """
        Identify the contents of file `filename`
        raises IOError if the file does not exist
        """
        self._thread_check()
        if not os.path.exists(filename):
            raise IOError("File does not exist: " + filename)
        try:
            return magic_file(self.cookie, filename)
        except MagicException as e:
            return self._handle509Bug(e)

    def _handle509Bug(self, e):
        # libmagic 5.09 has a bug where it might mail to identify the
        # mimetype of a file and returns null from magic_file (and
        # likely _buffer), but also does not return an error message.
        if e.message is None and (self.flags & MAGIC_MIME):
            return "application/octet-stream"

    def _thread_check(self):
        if self.thread != threading.currentThread():
            raise Exception('attempting to use libmagic on multiple threads will '
                            'end in SEGV.  Prefer to use the module functions '
                            'from_file or from_buffer, or carefully manage direct '
                            'use of the Magic class')

    def __del__(self):
        # no _thread_check here because there can be no other
        # references to this object at this point.

        # during shutdown magic_close may have been cleared already so
        # make sure it exists before using it.

        # the self.cookie check should be unnessary and was an
        # incorrect fix for a threading problem, however I'm leaving
        # it in because it's harmless and I'm slightly afraid to
        # remove it.
        if self.cookie and magic_close:
            magic_close(self.cookie)
            self.cookie = None


instances = threading.local()

def _get_magic_type(mime):
    i = instances.__dict__.get(mime)
    if i is None:
        i = instances.__dict__[mime] = Magic(mime=mime)
    return i

def from_file(filename, mime=False):
    """"
    Accepts a filename and returns the detected filetype.  Return
    value is the mimetype if mime=True, otherwise a human readable
    name.

    >>> magic.from_file("testdata/test.pdf", mime=True)
    'application/pdf'
    """
    m = _get_magic_type(mime)
    return m.from_file(filename)

def from_buffer(buffer, mime=False):
    """
    Accepts a binary string and returns the detected filetype.  Return
    value is the mimetype if mime=True, otherwise a human readable
    name.

    >>> magic.from_buffer(open("testdata/test.pdf").read(1024))
    'PDF document, version 1.2'
    """
    m = _get_magic_type(mime)
    return m.from_buffer(buffer)




libmagic = None
# Let's try to find magic or magic1
dll = ctypes.util.find_library('magic') or ctypes.util.find_library('magic1') or ctypes.util.find_library('cygmagic-1')

# This is necessary because find_library returns None if it doesn't find the library
if dll:
    libmagic = ctypes.CDLL(dll)

if not libmagic or not libmagic._name:
    platform_to_lib = {'darwin': ['/opt/local/lib/libmagic.dylib',
                                  '/usr/local/lib/libmagic.dylib'] +
                         # Assumes there will only be one version installed
                         glob.glob('/usr/local/Cellar/libmagic/*/lib/libmagic.dylib'),
                       'win32':  ['magic1.dll','cygmagic-1.dll']}
    for dll in platform_to_lib.get(sys.platform, []):
        try:
            libmagic = ctypes.CDLL(dll)
            break
        except OSError:
            pass

if not libmagic or not libmagic._name:
    # It is better to raise an ImportError since we are importing magic module
    raise ImportError('failed to find libmagic.  Check your installation')

magic_t = ctypes.c_void_p

def errorcheck_null(result, func, args):
    if result is None:
        err = magic_error(args[0])
        raise MagicException(err)
    else:
        return result

def errorcheck_negative_one(result, func, args):
    if result is -1:
        err = magic_error(args[0])
        raise MagicException(err)
    else:
        return result


def coerce_filename(filename):
    if filename is None:
        return None

    # ctypes will implicitly convert unicode strings to bytes with
    # .encode('ascii').  If you use the filesystem encoding 
    # then you'll get inconsistent behavior (crashes) depending on the user's
    # LANG environment variable
    is_unicode = (sys.version_info[0] <= 2 and
                  isinstance(filename, str)) or \
                  (sys.version_info[0] >= 3 and
                   isinstance(filename, str))
    if is_unicode:
        return filename.encode('utf-8')
    else:
        return filename

magic_open = libmagic.magic_open
magic_open.restype = magic_t
magic_open.argtypes = [c_int]

magic_close = libmagic.magic_close
magic_close.restype = None
magic_close.argtypes = [magic_t]

magic_error = libmagic.magic_error
magic_error.restype = c_char_p
magic_error.argtypes = [magic_t]

magic_errno = libmagic.magic_errno
magic_errno.restype = c_int
magic_errno.argtypes = [magic_t]

_magic_file = libmagic.magic_file
_magic_file.restype = c_char_p
_magic_file.argtypes = [magic_t, c_char_p]
_magic_file.errcheck = errorcheck_null

def magic_file(cookie, filename):
    return _magic_file(cookie, coerce_filename(filename))

_magic_buffer = libmagic.magic_buffer
_magic_buffer.restype = c_char_p
_magic_buffer.argtypes = [magic_t, c_void_p, c_size_t]
_magic_buffer.errcheck = errorcheck_null

def magic_buffer(cookie, buf):
    return _magic_buffer(cookie, buf, len(buf))


_magic_load = libmagic.magic_load
_magic_load.restype = c_int
_magic_load.argtypes = [magic_t, c_char_p]
_magic_load.errcheck = errorcheck_negative_one

def magic_load(cookie, filename):
    return _magic_load(cookie, coerce_filename(filename))

magic_setflags = libmagic.magic_setflags
magic_setflags.restype = c_int
magic_setflags.argtypes = [magic_t, c_int]

magic_check = libmagic.magic_check
magic_check.restype = c_int
magic_check.argtypes = [magic_t, c_char_p]

magic_compile = libmagic.magic_compile
magic_compile.restype = c_int
magic_compile.argtypes = [magic_t, c_char_p]



MAGIC_NONE = 0x000000 # No flags

MAGIC_DEBUG = 0x000001 # Turn on debugging

MAGIC_SYMLINK = 0x000002 # Follow symlinks

MAGIC_COMPRESS = 0x000004 # Check inside compressed files

MAGIC_DEVICES = 0x000008 # Look at the contents of devices

MAGIC_MIME = 0x000010 # Return a mime string

MAGIC_MIME_ENCODING = 0x000400 # Return the MIME encoding

MAGIC_CONTINUE = 0x000020 # Return all matches

MAGIC_CHECK = 0x000040 # Print warnings to stderr

MAGIC_PRESERVE_ATIME = 0x000080 # Restore access time on exit

MAGIC_RAW = 0x000100 # Don't translate unprintable chars

MAGIC_ERROR = 0x000200 # Handle ENOENT etc as real errors

MAGIC_NO_CHECK_COMPRESS = 0x001000 # Don't check for compressed files

MAGIC_NO_CHECK_TAR = 0x002000 # Don't check for tar files

MAGIC_NO_CHECK_SOFT = 0x004000 # Don't check magic entries

MAGIC_NO_CHECK_APPTYPE = 0x008000 # Don't check application type

MAGIC_NO_CHECK_ELF = 0x010000 # Don't check for elf details

MAGIC_NO_CHECK_ASCII = 0x020000 # Don't check for ascii files

MAGIC_NO_CHECK_TROFF = 0x040000 # Don't check ascii/troff

MAGIC_NO_CHECK_FORTRAN = 0x080000 # Don't check ascii/fortran

MAGIC_NO_CHECK_TOKENS = 0x100000 # Don't check ascii/tokens



# Python imports
import fnmatch
import logging
import os
import re
import datetime
import traceback
import io

# Project imports
from classified.incremental import Incremental
from classified.meta import Path, File, CorruptionError
from classified.probe import get_probe
from classified.report import get_report
import classified.report.all

# Third party imports
try:
    import markdown
except ImportError:
    markdown = None


class Scanner(object):
    def __init__(self, config, option):
        self.config = config
        self.option = option
        self.started = datetime.datetime.now()

        # Excluded file system types
        self.exclude_dirs = []
        try:
            self.exclude_dirs = self.config.getmulti('scanner', 'exclude_dirs')
        except self.config.Error:
            pass

        self.exclude_fs = []
        try:
            self.exclude_fs = self.config.getmulti('scanner', 'exclude_fs')
            self.exclude_fs = [re.compile(fnmatch.translate(pattern)) for pattern in self.exclude_fs]
        except self.config.Error:
            pass

        self.exclude_link = True
        try:
            self.exclude_link = self.config.getboolean('scanner', 'exclude_link')
        except self.config.Error:
            pass

        self.exclude_type = []
        try:
            self.exclude_type = self.config.getmulti('scanner', 'exclude_type')
            self.exclude_type = [re.compile(fnmatch.translate(pattern)) for pattern in self.exclude_type]
        except self.config.Error:
            pass

        self.exclude_repo = {}
        try:
            for item in self.config.getmulti('scanner', 'exclude_repo'):
                try:
                    repository_type, pattern = item.split(':', 1)
                except ValueError:
                    logging.error(
                        '%s: exclude_repo must have format "type:path"' % (
                            self.config.filename,
                        )
                    )
                    raise
                if repository_type not in self.exclude_repo:
                    self.exclude_repo[repository_type] = []
                self.exclude_repo[repository_type].append(re.compile(
                    fnmatch.translate(pattern)
                ))
        except self.config.Error:
            pass

        # Max traversal depths
        self.mindepth = int(self.config.getdefault('scanner', 'mindepth', -1))
        self.maxdepth = int(self.config.getdefault('scanner', 'maxdepth', -1))

        # Deflation of archives enabled?
        try:
            self.deflate = self.config.getboolean('scanner', 'deflate')
        except self.config.Error:
            self.deflate = True

        try:
            self.deflate_limit = self.config.getint('scanner', 'deflate_limit')
        except self.config.Error:
            self.deflate_limit = 5

        # Incremental enabled?
        try:
            if self.config.getboolean('scanner', 'incremental'):
                self.incremental = Incremental(self.config)
            else:
                self.incremental = False
        except self.config.Error:
            self.incremental = False

        # Report enabled?
        self.report = get_report(self.option.report_format, self.config,
                                 self.option)

        # Import probes
        probes = set(self.option.probes.split(','))
        try:
            probes.update(self.config.getlist('scanner', 'include_probe'))
        except self.config.Error:
            pass

        if 'all' in probes:
            # Remove it
            probes.remove('all')

            # Auto load all probe module names
            import classified.probe.all
            for probe in classified.probe.all.__all__:
                probe = getattr(classified.probe.all, probe)
                probes.add(probe.__module__.split('.')[-1])

        elif not probes:
            raise TypeError('No probes enabled, check -p')

        modules = {}
        for probe in probes:
            try:
                modules[probe] = __import__('classified.probe.%s' % probe)
            except ImportError as e:
                raise TypeError('Invalid probe %s enabled: %s' % (probe,
                    str(e)))

        # Setup probes
        self.probes = {}
        if self.config.has_section('probe'):
            for option in self.config.options('probe'):
                pattern = re.compile(fnmatch.translate(option))
                self.probes[pattern] = self.config.getlist('probe', option)

        else:
            for name in probes:
                probe = get_probe(name, self.config, self.report)
                for option in probe.target:
                    pattern = re.compile(fnmatch.translate(option))
                    if pattern not in self.probes:
                        self.probes[pattern] = []
                    self.probes[pattern].append(name)

    def probe(self, item, name):
        logging.debug('probe %s on %r' % (name, item))
        try:
            probe = get_probe(name, self.config, self.report)
            if probe.can_probe(item):
                probe.probe(item)
        except NotImplementedError:
            logging.warning('could not start probe %s: not implemented' % name)
        except Exception as error:
            logging.error('probe %s on %r failed: %s' % (name, item, error))

    def scan(self, path, max_depth=10):
        if os.path.isdir(path):
            for item in Path(path).walk(
                    recurse=True,
                    max_depth=max_depth,
                    deflate=self.deflate,
                    deflate_limit=self.deflate_limit,
                ):
                if self.scan_item(item) is StopIteration:
                    break

        else:
            self.scan_item(File(path))

    def scan_item(self, item):
        if item is None:
            return

        # No readable file? Skip
        if not item.readable:
            logging.debug('skipping %s: no readable content' % item)
            return

        # No mime type? Skip
        elif item.mimetype is None:
            logging.debug('skipping %s: no mimetype' % item)
            return

        # Mime type exclusions
        elif self.test_exclude_mimetype(item):
            logging.debug('skipping %s: excluded %s mime type' % (item,
                item.mimetype))
            return

        # File system type exclusions
        elif self.test_exclude_fs(item):
            logging.info('skipping %s: excluded %s filesystem' % (item,
                item.mount.fs['type']))
            return

        # Repository exclusions
        elif self.test_exclude_repo(item):
            logging.debug('skipping %s: excluded %s repository' % (item,
                item.repository.type))
            return

        elif self.incremental and item in self.incremental:
            logging.debug('skipping %s: file in incremental cache' % item)
            return

        else:
            logging.debug('scanning %r' % item)

        success = True
        for pattern, probes in self.probes.items():
            if pattern.match(item.mimetype):
                for probe in probes:
                    try:
                        self.probe(item, probe)
                    except CorruptionError as e:
                        logging.error('probe %s on %s failed: %s' % (probe,
                            item, e))
                        success = False

        if self.incremental and success:
            self.incremental.add(item)

    def test_exclude_fs(self, item):
        return item.mount.fs['type'] in self.exclude_fs

    def test_exclude_mimetype(self, item):
        return item.mimetype in self.exclude_type

    def test_exclude_repo(self, item):
        if item.repository.type is not None:
            for repository_type in (item.repository.type, 'any'):
                if repository_type in self.exclude_repo:
                    for pattern in self.exclude_repo[repository_type]:
                        if pattern.match(str(item)):
                            return True

        return False



import logging
import sys
from classified.report.syslogger import SyslogReport


class FileReport(SyslogReport):
    name = 'file'
    format = '%(asctime)s %(message)s'

    def setup(self):
        if not self.option.output:
            print('Please supply an output file with --output', file=sys.stderr)
            sys.exit(1)

        # Setup logger environment
        self.logger = logging.getLogger('classified')
        self.handle = logging.FileHandler(self.option.output)
        formatter = logging.Formatter(
            self.config.getdefault('report:file',
                                   'format',
                                   self.format)
        )
        self.handle.setFormatter(formatter)
        self.logger.addHandler(self.handle)
        self.logger.setLevel(logging.INFO)

    def emit(self, message):
        self.logger.info(message)



# Python imports
import stat

# Project imports
from classified.probe.base import Probe


class SSL(Probe):
    target = ('text/*',)
    format = '{filename}[{line:d}]: {key_info} {key_type} {username}'

    def probe(self, item):
        item.open()
        try:
            key = False
            key_info = ['plaintext']
            key_type = None

            data = b''
            line = b''
            lineno = 0
            while line == '':
                try:
                    data = item.readline()
                except Exception as e:
                    break

                line = data.strip()
                lineno += 1
                if data == '':  # EOF?
                    break

            if b'-----BEGIN RSA PRIVATE KEY-----' in line:
                key = 'RSA pri)ate key'
                key_type = 'rsa'
            elif b'SSH PRIVATE KEY FILE FORMAT 1' in line:
                key = 'RSA1 private key'
                key_type = 'rsa1'
            elif b'-----BEGIN DSA PRIVATE KEY-----' in line:
                key = 'DSA private key'
                key_type = 'dsa'
            elif b'-----BEGIN EC PRIVATE KEY-----' in line:
                key = 'ECDSA private key'
                key_type = 'ecdsa'
            else:
                # No SSH private key was found
                return

            data = ''
            line = ''
            while line == '':
                try:
                    data = item.readline()
                except Exception as e:
                    return
                line = data.strip()

            if line.startswith('Proc-Type:') and 'ENCRYPTED' in line:
                key_info = ['encrypted']

            info = item.stat()
            mode = info[stat.ST_MODE]
            # Check if file is group or world-readable
            if mode & stat.S_IRGRP or mode & stat.S_IROTH:
                key_info.append('world-readable')
            else:
                key_info.append('protected')

            self.record(item,
                raw=data,
                line=lineno,
                key=key,
                key_info=' '.join(key_info),
                key_type=key_type,
            )

        finally:
            item.close()



# Python imports
import hashlib
import struct
from zlib import adler32, crc32


class Adler32(object):
    block_size = 64
    digestsize = 4
    digest_size = 4
    name = 'adler32'

    def __init__(self, string=''):
        self.checksum = adler32(string)

    def update(self, string):
        self.checksum = adler32(string, self.checksum)

    def digest(self):
        return struct.pack('>L', self.checksum & 0xffffffff)

    def hexdigest(self):
        return '%08x' % (self.checksum & 0xffffffff,)


class CRC32(object):
    block_size = 64
    digestsize = 4
    digest_size = 4
    name = 'crc32'

    def __init__(self, string=''):
        self.checksum = crc32(string)

    def update(self, string):
        self.checksum = crc32(string, self.checksum)

    def digest(self):
        return struct.pack('>L', self.checksum & 0xffffffff)

    def hexdigest(self):
        return '%08x' % (self.checksum & 0xffffffff,)


def new(algorithm, string=''):
    if algorithm == 'adler32':
        return Adler32(string)

    if algorithm == 'crc32':
        return CRC32(string)

    else:
        return hashlib.new(algorithm)



# Python imports
from collections import defaultdict
import datetime
import socket
import sys

# Project imports
from classified.report.base import Report


class HTMLReport(Report):
    name = 'html'
    template = 'report/full.html'

    def setup(self):
        if not self.option.output:
            print('Please supply an output file with --output', file=sys.stderr)
            sys.exit(1)

        # Setup template environment
        self.setup_env()

    def setup_env(self):
        # Load Jinja2
        from jinja2 import Environment, PackageLoader

        # Load template file
        self.env = Environment(loader=PackageLoader('classified', 'template'))
        self.filename = self.config.getdefault('report:{}'.format(self.name),
                                               'template',
                                               self.template)
        self.template = self.env.get_template(self.filename)

        # Collected entries
        self.entries = dict(
            fqdn=socket.getfqdn(),
            hostname=socket.gethostname(),
            time=dict(start=datetime.datetime.now(), finish=None),
            probe=defaultdict(list),
            filename=defaultdict(int),
            username=defaultdict(int),
        )


    def report(self, probe, item, **kwargs):
        self.entries['probe'][probe.name].append((item, kwargs))
        self.entries['filename'][str(item)] += 1
        if 'username' in kwargs:
            self.entries['username'][kwargs['username']] += 1
        elif 'uid' in kwargs:
            self.entries['username'][kwargs['uid']] += 1

    def render(self):
        if self.option.output == '-':
            fp = sys.stdout
        else:
            fp = open(self.option.output, 'w')

        self.entries['time']['finish'] = datetime.datetime.now()
        fp.write(self.template.render(**self.entries))



# Python imports
import os
import subprocess
import sys
try:
    import win32com.client
except ImportError:
    pass


def get_filesystem(path, filesystems=None):
    '''
    Get the filesystem of a given path.
    '''
    path = os.path.abspath(path)
    if filesystems is None:
        filesystems = get_filesystems()

    filesystems_match = []
    for filesystem in filesystems:
        if path.startswith(filesystem['mount']):
            filesystems_match.append(filesystem)

    # Sort by longest matching path
    def _mount(filesystem):
        return filesystem['mount']

    filesystems_match.sort(key=_mount)
    return filesystems_match[0]


def get_filesystems():
    '''List mounted file systems.'''
    if sys.platform.startswith('linux'):
        filesystems = _get_filesystems_linux()
    elif sys.platform in ('darwin', 'freebsd', 'openbsd', 'netbsd'):
        filesystems = _get_filesystems_bsd()
    elif sys.platform.startswith('win'):
        filesystems = _get_filesystems_windows()
    else:
        filesystems = []
    return filesystems


def _get_filesystems_linux():
    '''List mounted file systems on Linux.'''
    with open('/etc/mtab', 'r') as handle:
        for line in handle:
            part = line.split()
            if len(part) < 6:
                continue
            yield {
                'device': part[0],
                'mount': part[1],
                'type': part[2],
                'options': part[3].split(','),
            }


def _get_filesystems_bsd():
    '''List mounted file systems on BSD.'''
    mount = subprocess.run('mount', capture_output=True, check=True)
    for line in mount.stdout.splitlines():
        part = line.decode('ascii').split()
        if len(part) < 4:
            continue
        yield {
            'device': part[0],
            'mount': part[2],
            'type': part[3].lstrip('(').rstrip(','),
        }


def _get_filesystems_windows():
    '''List mounted drives on Window.'''
    wmi_service = win32com.client.Dispatch('WbemScripting.SWbemLocator')
    services = wmi_service.ConnectServer('.', "root\\cimv2")
    items = services.ExecQuery('Select * from Win32_LogicalDisk')
    for item in items:
        yield {
            'device': item.DeviceID,
            'mount': item.VolumeName,
        }

if __name__ == '__main__':
    filesystems = list(get_filesystems())
    print('{} file systems mounted:'.format(len(filesystems)))
    for filesystem in filesystems:
        print('  {device} on {mount}'.format(**filesystem))

    print('filesystem of /:')
    print('  ', get_filesystem('/'))




import datetime
import os
from classified.report.html import HTMLReport


class TTYReport(HTMLReport):
    name = 'tty'
    template = 'report/full.tty'

    def setup(self):
        # Setup template environment
        self.setup_env()

    def render(self):
        self.entries['user'] = os.environ.get('USER', 'no-reply')
        self.entries['time']['finish'] = datetime.datetime.now()

        # Compile MIME message
        print(self.template.render(**self.entries))



# Python imports
from configparser import ConfigParser, Error, NoOptionError, NoSectionError
import os
import re


class Config(ConfigParser):
    Error = Error
    NoOptionError = NoOptionError
    NoSectionError = NoSectionError

    re_padding = re.compile(r'(^[\s\t,]+|[\s\t,]+$)')

    def __init__(self, filename, *args, **kwargs):
        ConfigParser.__init__(self, *args, **kwargs)
        if filename:
            if os.access(filename, os.R_OK):
                self.read([filename])
            else:
                raise IOError('Could not open {} for reading'.format(filename))

    def getdefault(self, section, option, default=None):
        try:
            return self.get(section, option)
        except (NoOptionError, NoSectionError):
            return default

    def getlist(self, section, option, sep=','):
        value = self.get(section, option)
        return [item.strip() for item in value.split(sep)]

    def getmulti(self, section, option, strip_comments=True):
        multi = []
        for item in self.get(section, option).splitlines():
            # Remove comments, if enabled
            if strip_comments:
                item = item.rsplit(' #', 1)[0]

            # Clean up padding
            item = self.re_padding.sub('', item)
            multi.append(item)
        return multi

    def read(self, filename):
        if isinstance(filename, (list, tuple)):
            self.filename = filename[0]
        else:
            self.filename = filename

        ConfigParser.read(self, filename)



# Python imports
import fnmatch
import grp
import logging
import os
import pwd
import re
import stat
import sys

# Project imports
from classified import checksum
from classified.probe import PROBES, IGNORE


class ProbeTracker(type):
    def __new__(cls, name, bases, attrs):
        new = type.__new__(cls, name, bases, attrs)
        if getattr(new, 'name', None) is None:
            PROBES[name.lower()] = new
        else:
            PROBES[new.name] = new
        return new


class Probe(object, metaclass=ProbeTracker):
    default_buffer = sys.stdout
    target = ()                     # default list of target mime types
    format = None                   # default format string for reporting
    name = None

    def __init__(self, config, report, buffer=None):
        self.config = config
        self.report = report
        self.name = self.name or self.__class__.__name__.lower()

        # Lookup algorithm for doing checks
        self.algorithm = self.config.getdefault('clean:%s' % self.name,
            'algorithm', self.config.getdefault('clean', 'algorithm', 'sha1')
        )

        if buffer is None:
            self.buffer = self.default_buffer
        else:
            self.buffer = buffer

        # See if the ignores are already parsed
        if self.name not in IGNORE:
            IGNORE[self.name] = dict(name=[], hash=[])

            # Ignored hashes
            try:
                ignore_hash = self.config.getmulti('clean:%s' % self.name,
                    'ignore_hash')
                IGNORE[self.name]['hash'] = ignore_hash
            except (self.config.NoOptionError, self.config.NoSectionError):
                IGNORE[self.name]['hash'] = []

            # Ignored names
            try:
                ignore_name = self.config.getmulti('clean:%s' % self.name,
                    'ignore_name')
                IGNORE[self.name]['name'] = [re.compile(fnmatch.translate(pattern)) for pattern in ignore_name]
            except (self.config.NoOptionError, self.config.NoSectionError):
                IGNORE[self.name]['name'] = []

            # Ignored repos
            try:
                ignore_repo = self.config.getmulti('clean:%s' % self.name,
                    'ignore_repo')
                IGNORE[self.name]['repo'] = []
                for ignore in ignore_repo:
                    repo_type, pattern = ignore.split(':', 1)
                    IGNORE[self.name]['repo'].append((
                        repo_type, re.compile(fnmatch.translate(pattern))
                    ))
            except (self.config.NoOptionError, self.config.NoSectionError):
                IGNORE[self.name]['repo'] = []

    def __unicode__(self):
        return self.name

    def can_probe(self, item):
        '''
        Tests if this probe can be ran against the given item.
        '''
        for hook in (self.ignore_name, self.ignore_repo):
            if hook(item):
                logging.debug('ignoring %r in %s: in %s' % (
                    item, self.name, hook.__name__
                ))
                return False

        # We're good, file is not ignored
        return True

    def ignore_hash(self, item, **kwargs):
        '''
        Check if the match is to be ignored, based on hash. This mechanism is
        used to filter out errors and false positives.
        '''
        context = self.config.getdefault('clean:%s' % self.name, 'context',
            self.config.getdefault('clean', 'context', 'line')
        )

        # Calculate checksum using the selected algorith, and check if it is in
        # the ignore list for this probe
        hashing = checksum.new(self.algorithm)
        if context == 'file':
            for line in item.open('rb'):
                hashing.update(line)

        elif context == 'line':
            try:
                if isinstance(kwargs['raw'], str):
                    hashing.update(kwargs['raw'].encode('utf-8'))
                else:
                    hashing.update(kwargs['raw'])
            except KeyError:
                # The reported item has no "raw" format, therefor we can not
                # provide a line-based hash
                return None, False

        elif context == 'format':
            format = self.config.get('clean:%s' % self.name, 'format')
            hashing.update(format.format(**kwargs))

        else:
            raise TypeError('Probe %s does not support %s context' % \
                (self.name, context))

        digest = hashing.hexdigest()
        if digest in IGNORE[self.name]['hash']:
            logging.debug('ignoring %r in %s: %s' % (item, self.name, digest))
            return digest, True
        else:
            logging.debug('allowing %r in %s: %s' % (item, self.name, digest))
            return digest, False

    def ignore_name(self, item):
        '''
        Check if the full path is to be ignored by this type of probe.
        '''
        for pattern in IGNORE[self.name]['name']:
            if pattern.match(str(item)):
                return True
        return False

    def ignore_repo(self, item):
        '''
        Check if the full path is to be ignored by this type of probe.
        '''
        if item.repository.type is None:
            return False

        for repository_type, pattern in IGNORE[self.name]['repo']:
            if repository_type not in (item.repository.type, 'any'):
                continue
            if pattern.match(str(item)):
                return True

        return False

    def probe(self, item):
        raise NotImplementedError

    def record(self, item, **kwargs):
        # Check if we can/may report this item
        digest, ignore = self.ignore_hash(item, **kwargs)
        if ignore:
            return

        # Exend kwargs
        kwargs['hash'] = digest
        kwargs['filename'] = str(item)
        kwargs['filename_relative'] = str(item).replace(os.getcwd(), '.')

        # Find out who owns the file
        info = item.stat()
        kwargs['uid'] = info[stat.ST_UID]
        kwargs['gid'] = info[stat.ST_GID]
        try:
            kwargs['username'] = pwd.getpwuid(kwargs['uid'])[0]
        except KeyError:
            kwargs['username'] = str(kwargs['uid'])
        try:
            kwargs['group'] = grp.getgrgid(kwargs['gid'])[0]
        except KeyError:
            kwargs['group'] = str(kwargs['gid'])

        # Send findings to reporting engine
        self.report.report(self, item, **kwargs)


def isdigit(x):
    if isinstance(x, str):
        return x.isdigit()
    return ord('0') <= x and x <= ord('9')



import syslog
from classified.report.base import Report


class SyslogReport(Report):
    name = 'syslog'

    def setup(self):
        # Open syslog
        facility = getattr(syslog, 'LOG_{}'.format(
            self.config.getdefault('report:syslog', 'syslog_facility',
                                   'daemon').upper()
        ))
        syslog.openlog('classified',
                       syslog.LOG_PID,
                       facility)

    def report(self, probe, item, **kwargs):
        try:
            formatter = self.config.get('report:syslog',
                                        'format_{}'.format(probe.name))
        except self.config.Error:
            formatter = probe.format
        message = formatter.format(**kwargs)
        self.emit(message)

    def emit(self, message):
        syslog.syslog(message)

    def render(self, *args, **kwargs):
        pass  # No rendering step for syslog



# Probe lookup map
PROBES = dict()
# Probe lookup map for ignores
IGNORE = dict()
# Report counters
REPORT = dict()


def get_probe(name, *args, **kwargs):
    try:
        return PROBES[name](*args, **kwargs)
    except KeyError:
        raise NotImplementedError(name)



# Python imports
import struct

# Project imports 
from classified.probe.base import Probe


class PCAP(Probe):
    magic = 0xa1b2c3d4
    target = ('application/octet-stream', 'application/vnd.tcpdump.pcap')
    format = '{filename_relative}[{line:d}]: pcap v{version} ({linktype})'
    header = 'IHHiIII'
    header_size = struct.calcsize(header)
    linktype = {
        0:   'NULL',
        1:   'Ethernet',
        3:   'AX25',
        6:   'IEEE802.5',
        7:   'ARCNet BSD',
        8:   'SLIP',
        9:   'PPP',
        10:  'FDDI',
        50:  'PPP HDLC',
        51:  'PPP ETHER',
        100: 'ATM RFC1483',
        101: 'RAW',
        104: 'C_HDLC',
        105: 'IEEE802.11',
        107: 'FRELAY',
        108: 'LOOP',
        113: 'LINUX SLL',
        114: 'LTALK',
        117: 'PFLOG',
        119: 'IEEE802.11 PRISM',
        122: 'IP over FC',
        123: 'SUNATM',
        127: 'IEEE802.11 RADIOTAP',
        129: 'ARCNET Linux',
        138: 'Apple IP over IEEE1394',
        139: 'MTP2 with PHDR',
        140: 'MTP2',
        141: 'MTP3',
        142: 'SCCP',
        143: 'DOCSIS',
        144: 'Linux IRDA',
        163: 'IEEE802.11 AVS',
        165: 'BACNET MS TP',
        166: 'PPP PPPD',
        169: 'GPRS LLC',
        177: 'Linux LAPD',
        187: 'Bluetooth HCI H4',
        189: 'USB Linux',
        192: 'PPI',
        195: 'IEEE802.15-4',
        196: 'SITA',
        197: 'ERF',
        201: 'Bluetooth HCI H4 with PHDR',
        202: 'AX25 KISS',
        203: 'LAPD',
        204: 'PPP with DIR',
        205: 'C_HDLC with DIR',
        206: 'FRELAY with DIR',
        209: 'IPMB Linux',
        215: 'IEEE802.15-4 NONASK PHY',
        220: 'USB Linux mmapped',
        224: 'FC 2',
        225: 'FC 2 with frame delims',
        226: 'IPNET',
        227: 'CAN SOCKETCAN',
        228: 'IPv4',
        229: 'IPv6',
        230: 'IEEE802.15-4 NOFCS',
        231: 'DBUS',
        235: 'DVB CI',
        236: 'MUX27010',
        237: 'STANAG 5066-D PDU',
        239: 'NFLOG',
        240: 'Netanalyzer',
        241: 'Netanalyzer Transparent',
        242: 'IPOIB',
        243: 'MPEG-2 TS',
        244: 'NG40',
        245: 'NFC LLCP',
        247: 'Infiniband',
        248: 'SCTP',
    }

    def probe(self, item):
        item.open('rb')
        chunk = item.read(self.header_size)
        if len(chunk) != self.header_size:
            return

        magic, version_major, version_minor, thiszone, \
            sigfigs, snaplen, network = struct.unpack(self.header, chunk)

        if magic == self.magic:
            self.record(item,
                line=1,
                version='%d.%d' % (version_major, version_minor),
                version_major=version_major,
                version_minor=version_minor,
                linktype=self.linktype.get(network, 'Unknown'),
            )



# Python imports
import datetime
import os
import smtplib
import sys
from email.mime.text import MIMEText

# Project imports
from classified.report.html import HTMLReport


class MailReport(HTMLReport):
    name = 'mail'
    sender = '{user}@{hostname}'
    subject = 'Classified report for {fqdn}'
    template = 'report/full.eml'

    def setup(self):
        if not self.option.output:
            print('Please supply recipients with --output', file=sys.stderr)
            sys.exit(1)

        # Setup template environment
        self.setup_env()

    def render(self):
        self.entries['user'] = os.environ.get('USER', 'no-reply')
        self.entries['time']['finish'] = datetime.datetime.now()

        subject = self.config.getdefault('report:mail', 'subject',
                                         self.subject)
        sender = self.config.getdefault('report:mail', 'sender',
                                        self.sender)

        # Compile MIME message
        message = MIMEText(self.template.render(**self.entries))
        message['Subject'] = subject.format(**self.entries)
        message['From'] = sender.format(**self.entries)
        message['To'] = self.option.output
        message['X-Mailer'] = 'classified/1.0'

        print(message.as_string())

        # Send mail
        smtp = smtplib.SMTP(self.config.getdefault('report:mail',
                                                   'server',
                                                   'localhost'))
        smtp.set_debuglevel(9)
        smtp.sendmail(sender.format(**self.entries),
                      self.option.output.split(','),
                      message.as_string())
        try:
            smtp.quit()
        except:
            pass



from distutils.core import setup


setup(
    name         = 'classified',
    version      = '1.4.0',
    author       = 'Wijnand Modderman',
    author_email = 'maze@pyth0n.org',
    description  = 'Classified data scanner',
    license      = 'MIT',
    keywords     = 'classified sensitive pan pci',
    packages     = [
        'classified',
        'classified.probe',
        'classified.probe.pan',
        'classified.probe.password',
        'classified.probe.pcap',
        'classified.probe.ssl',
        'classified.report',
    ],
    package_data = {
        'classified': ['template/*/*'],
    },
    data_files   = [
        ('/etc/classified', ['etc/classified.conf.example']),
    ],
    scripts      = ['bin/classified'],
)



# Project imports
from classified.report import REPORT


class ReportTracker(type):
    def __new__(cls, name, bases, attrs):
        new = type.__new__(cls, name, bases, attrs)
        if getattr(new, 'name', None) is None:
            REPORT[name.lower()] = new
        else:
            REPORT[new.name] = new
        return new


class Report(object, metaclass=ReportTracker):
    def __init__(self, config, option):
        self.config = config
        self.option = option
        self.setup()

    def setup(self):
        pass

    def report(self, probe, item, **kwargs):
        raise NotImplementedError

    def render(self):
        raise NotImplementedError

    # Alias
    __call__ = report



# Python imports
try:
    import dbm.ndbm
except ImportError:
    import dbm as dbm
import logging

# Project imports
from classified import checksum


class Incremental(object):
    default_algorithm = 'sha1'
    default_blocksize = 16384

    def __init__(self, config):
        self.config = config

        # Local cache
        self.cache = {}

        # Configuration bits
        self.algorithm = self.config.getdefault('incremental', 'algorithm',
            self.default_algorithm)
        self.database = self.config.get('incremental', 'database')
        try:
            self.blocksize = self.config.getint('incremental', 'blocksize')
        except self.config.NoOptionError:
            self.blocksize = self.default_blocksize

        # Open the database in secure mode
        self.db = dbm.ndbm.open(self.database, 'c', 0o600)

        logging.info('only checking incremental changes')
        logging.debug('tracking incremental changes in %s' % self.database)

    def __contains__(self, item):
        if str(item) in self.db:
            old_value = self.db[str(item)]
            new_value = self.cache.get(item, self.checksum(item))
            return old_value == new_value
        else:
            return False

    def add(self, item):
        self.db[str(item)] = self.cache.get(item, self.checksum(item))

    def checksum(self, item):
        if self.algorithm == 'mtime':
            return str(int(item.mtime))

        else:
            method = checksum.new(self.algorithm)
            handle = item.open()
            while True:
                chunk = handle.read(self.blocksize)
                if not chunk:
                    break
                else:
                    method.update(chunk)
            handle.close()

            self.cache[item] = method.hexdigest()
            return self.cache[item]



# Parts are courtesey of `Ben Hogdson <http://benhodgson.com/>`_.

# Python imports
import logging
import re

# Project imports
from classified.probe.base import Probe, isdigit


decimal_decoder = lambda s: int(s, 10)
decimal_encoder = lambda i: str(i)


def luhn_sum_mod_base(string, base=10, decoder=decimal_decoder):
    # Adapted from http://en.wikipedia.org/wiki/Luhn_algorithm
    digits = list(map(decoder, string))
    return (sum(digits[::-2]) +
        sum([sum(divmod(2*d, base)) for d in digits[-2::-2]])) % base


def generate(string, base=10, encoder=decimal_encoder,
    decoder=decimal_decoder):
    '''
    Calculates the Luhn mod N check character for the given input string. This
    character should be appended to the input string to produce a valid Luhn
    mod N string in the given base.

    >>> value = '4205092350249'
    >>> generate(value)
    '1'

    When operating in a base other than decimal, encoder and decoder callables
    should be supplied. The encoder should take a single argument, an integer,
    and return the character corresponding to that integer in the operating
    base. Conversely, the decoder should take a string containing a single
    character and return its integer value in the operating base. Note that
    the mapping between values and characters defined by the encoder and
    decoder should be one-to-one.

    For example, when working in hexadecimal:

    >>> hex_alphabet = '0123456789abcdef'
    >>> hex_encoder = lambda i: hex_alphabet[i]
    >>> hex_decoder = lambda s: hex_alphabet.index(s)
    >>> value = 'a8b56f'
    >>> generate(value, base=16, encoder=hex_encoder, decoder=hex_decoder)
    'b'
    >>> verify('a8b56fb', base=16, decoder=hex_decoder)
    True
    >>> verify('a8b56fc', base=16, decoder=hex_decoder)
    False
    '''

    d = luhn_sum_mod_base(string+encoder(0), base=base, decoder=decoder)
    if d != 0:
        d = base - d
    return encoder(d)


def verify(string, base=10, decoder=decimal_decoder):
    '''
    Verifies that the given string is a valid Luhn mod N string.

    >>> verify('5105105105105100') # MasterCard test number
    True

    When operating in a base other than decimal, encoder and decoder callables
    should be supplied. The encoder should take a single argument, an integer,
    and return the character corresponding to that integer in the operating
    base. Conversely, the decoder should take a string containing a single
    character and return its integer value in the operating base. Note that
    the mapping between values and characters defined by the encoder and
    decoder should be one-to-one.

    For example, 'b' is the correct check character for the hexadecimal string
    'a8b56f':

    >>> hex_decoder = lambda s: '0123456789abcdef'.index(s)
    >>> verify('a8b56fb', base=16, decoder=hex_decoder)
    True

    Any other check digit (in this example: 'c'), will result in a failed
    verification:

    >>> verify('a8b56fc', base=16, decoder=hex_decoder)
    False
    '''
    return luhn_sum_mod_base(string, base=base, decoder=decoder) == 0


def mask(card_number, keep=4):
    '''
    Mask a card number so it's suitable for printing.
    '''
    keep *= -1
    return '*' * len(card_number[:keep]) + card_number[keep:]


class PAN(Probe):
    '''
    Scan for Primary Account Number (PAN) data in (text) files.
    '''

    target = ('text/*',)
    format = '{filename}[{line:d}]: {company} {card_number_masked}'
    ignore = list('\x00-:\r\n')
    _check = {
        'American Express': dict(
            length = [15],
            prefix = re.compile(r'^3[47]'),
        ),
        'Diners Club EnRoute': dict(
            length = [15],
            prefix = re.compile(r'^(?:2014|2149)'),
        ),
        'Diners Club Carte Blanche': dict(
            length = [14],
            prefix = re.compile(r'^30[1-5]'),
        ),
        'Diners Club International': dict(
            length = [14],
            prefix = re.compile(r'^36'),
        ),
        'Diners Club America': dict(
            length = [14],
            prefix = re.compile(r'^5[45]'),
        ),
        'Discover': dict(
            length = [16],
            prefix = re.compile(r'^6011'),
        ),
        'InstaPayment': dict(
            length = [16],
            prefix = re.compile(r'^63[7-9]'),
        ),
        'JCB': dict(
            length = [16],
            prefix = re.compile(r'^(?:3088|3096|3112|3158|3337|352[89]|35[3-7][0-9]|358[0-9])'),
        ),
        'Laser': dict(
            length = list(range(12, 20)),
            prefix = re.compile(r'^(?:6304|6706|6771|6709)'),
        ),
        'Maestro': dict(
            length = list(range(12, 20)),
            prefix = re.compile(r'^(?:5018|5020|5038|5893|6304|6759|676[1-3]|0604)'),
        ),
        'MasterCard': dict(
            length = [16],
            prefix = re.compile(r'^5[1-5]'),
        ),
        'VISA': dict(
            length = [13, 16],
            prefix = re.compile(r'^4'),
        ),
    }

    def __init__(self, config, *args, **kwargs):
        super(PAN, self).__init__(config, *args, **kwargs)

        # Also keep track of per prefix size checks
        self._check_size = {}
        for company, checks in self._check.items():
            for length in checks['length']:
                if length not in self._check_size:
                    self._check_size[length] = {}
                self._check_size[length][company] = checks['prefix']

        # Ignores, if configured
        if self.config.has_option('probe:pan', 'ignore'):
            self.ignore = [chr(int(char, 16)) for char in self.config.getlist('probe:pan', 'ignore')]

    def luhn_check(self, card_number):
        # Do the Luhn check
        if verify(card_number):
            return self.process_prefix(card_number)

    def process_prefix(self, card_number):
        length = len(card_number)
        if length in self._check_size:
            for company, prefix in self._check_size[length].items():
                if prefix.match(card_number):
                    return company

    def probe(self, item):
        # Keep track of consecutive ranges of numbers, stripping out potential
        # padding characters

        digits = []
        digits_min = min(self._check_size)
        digits_max = max(self._check_size)

        line = 0
        hits = 0
        try:
            limit = self.config.getint('probe:pan', 'limit')
        except self.config.Error:
            limit = 0

        prev = chr(0)
        for text in item.open():
            line += 1

            for char in text:
                # If we have a digit, append it to the digits list
                if isdigit(char):
                    digits.append(int(char))

                    if len(digits) >= digits_max:
                        digits = digits[1:]

                    if len(digits) >= digits_min:
                        for x in range(digits_min, digits_max + 1):
                            card_number = ''.join(map(str, digits[:x]))
                            card_company = self.luhn_check(card_number)
                            if card_company is not None:
                                self.record(item,
                                    raw=text,
                                    line=line,
                                    card_number=card_number,
                                    card_number_masked=mask(card_number),
                                    company=card_company,
                                )

                                # Rotate digits
                                digits = digits[x:]

                                # Keep track of hits
                                hits += 1
                                if limit and hits >= limit:
                                    logging.debug('pan probe hit limit '
                                                  'of %d' % limit)
                                    return
                                break

                # We ignore dashes, new lines and carriage returns
                elif char in self.ignore:
                    # .. if we have two successive ignored characters, reset
                    # the digits array
                    if prev in self.ignore:
                        digits = []

                # Otherwise we'll reset the buffer
                else:
                    digits = []

                # Keep track of the previous character
                prev = char



# Project imports
from classified.report.file import FileReport
from classified.report.html import HTMLReport
from classified.report.mail import MailReport
from classified.report.syslogger import SyslogReport
from classified.report.tty import TTYReport

__all__ = ['HTMLReport', 'MailReport', 'SyslogReport', 'TTYReport']
